@@ -46,7 +46,7 @@
 /* stuff from common/ */
 #include <gdaemonmount.h>
 #include <gvfsdaemonprotocol.h>
-#include <gdbusutils.h>
+#include <gvfsdbusutils.h>
 
 #define FUSE_USE_VERSION 26
 #include <fuse.h>
@@ -75,26 +75,26 @@ typedef struct {
   gchar    *path;
   FileOp    op;
   gpointer  stream;
-  gint      length;
-  size_t    pos;
+  goffset   pos;
 } FileHandle;
 
-static GThread       *subthread           = NULL;
-static GMainLoop     *subthread_main_loop = NULL;
-static GVfs          *gvfs                = NULL;
+static GThread        *subthread             = NULL;
+static GMainLoop      *subthread_main_loop   = NULL;
+static GVfs           *gvfs                  = NULL;
 
-static GVolumeMonitor *volume_monitor     = NULL;
+static GVolumeMonitor *volume_monitor        = NULL;
 
 /* Contains pointers to MountRecord */
-static GList         *mount_list          = NULL;
-static GMutex        *mount_list_mutex;
+static GList          *mount_list            = NULL;
+static GMutex         *mount_list_mutex;
 
-static time_t         daemon_creation_time;
-static uid_t          daemon_uid;
-static gid_t          daemon_gid;
+static time_t          daemon_creation_time;
+static uid_t           daemon_uid;
+static gid_t           daemon_gid;
 
-static GStaticMutex   global_mutex        = G_STATIC_MUTEX_INIT;
-static GHashTable    *global_fh_table     = NULL;
+static GStaticMutex    global_mutex          = G_STATIC_MUTEX_INIT;
+static GHashTable     *global_path_to_fh_map = NULL;
+static GHashTable     *global_active_fh_map  = NULL;
 
 /* ------- *
  * Helpers *
@@ -165,6 +165,7 @@ errno_from_error (GError *error)
       { G_IO_ERROR_WRONG_ETAG,         EIO          },
       { G_IO_ERROR_TIMED_OUT,          EIO          },
       { G_IO_ERROR_BUSY,               EBUSY        },
+      { G_IO_ERROR_WOULD_BLOCK,        EAGAIN       },
       { -1,                            -1           }
     };
 
@@ -183,6 +184,29 @@ errno_from_error (GError *error)
   return EIO;
 }
 
+/* Conveys the pid of the client to filesystem backends; see
+ * get_pid_for_file() in gdaemonfile.c
+ *
+ * can only be called during a filesystem op
+ */
+static void
+set_pid_for_file (GFile *file)
+{
+  struct fuse_context *context;
+
+  if (file == NULL)
+    goto out;
+
+  context = fuse_get_context ();
+  if (context == NULL)
+    goto out;
+
+  g_object_set_data (G_OBJECT (file), "gvfs-fuse-client-pid", GUINT_TO_POINTER (context->pid));
+
+ out:
+  ;
+}
+
 static FileHandle *
 file_handle_new (const gchar *path)
 {
@@ -194,6 +218,8 @@ file_handle_new (const gchar *path)
   file_handle->op = FILE_OP_NONE;
   file_handle->path = g_strdup (path);
 
+  g_hash_table_insert (global_active_fh_map, file_handle, file_handle);
+
   return file_handle;
 }
 
@@ -209,14 +235,18 @@ file_handle_unref (FileHandle *file_handle)
 {
   if (g_atomic_int_dec_and_test (&file_handle->refcount))
     {
+      gint refs;
+
       g_static_mutex_lock (&global_mutex);
 
       /* Test again, since e.g. get_file_handle_for_path() might have
        * snatched the global mutex and revived the file handle between
-       * g_atomic_int_dec_and_test() and us obtaining the lock. */
+       * g_atomic_int_dec_and_test() and us obtaining the global lock. */
 
-      if (g_atomic_int_get (&file_handle->refcount) == 0)
-        g_hash_table_remove (global_fh_table, file_handle->path);
+      refs = g_atomic_int_get (&file_handle->refcount);
+
+      if (refs == 0)
+        g_hash_table_remove (global_path_to_fh_map, file_handle->path);
 
       g_static_mutex_unlock (&global_mutex);
     }
@@ -252,6 +282,8 @@ file_handle_close_stream (FileHandle *file_handle)
 static void
 file_handle_free (FileHandle *file_handle)
 {
+  g_hash_table_remove (global_active_fh_map, file_handle);
+
   file_handle_close_stream (file_handle);
   g_mutex_free (file_handle->mutex);
   g_free (file_handle->path);
@@ -265,12 +297,12 @@ get_file_handle_for_path (const gchar *path)
 
   g_static_mutex_lock (&global_mutex);
 
-  fh = g_hash_table_lookup (global_fh_table, path);
+  fh = g_hash_table_lookup (global_path_to_fh_map, path);
+
   if (fh)
     file_handle_ref (fh);
 
   g_static_mutex_unlock (&global_mutex);
-
   return fh;
 }
 
@@ -281,7 +313,7 @@ get_or_create_file_handle_for_path (const gchar *path)
 
   g_static_mutex_lock (&global_mutex);
 
-  fh = g_hash_table_lookup (global_fh_table, path);
+  fh = g_hash_table_lookup (global_path_to_fh_map, path);
 
   if (fh)
     {
@@ -290,11 +322,30 @@ get_or_create_file_handle_for_path (const gchar *path)
   else
     {
       fh = file_handle_new (path);
-      g_hash_table_insert (global_fh_table, fh->path, fh);
+      g_hash_table_insert (global_path_to_fh_map, fh->path, fh);
     }
 
   g_static_mutex_unlock (&global_mutex);
+  return fh;
+}
 
+static FileHandle *
+get_file_handle_from_info (struct fuse_file_info *fi)
+{
+  FileHandle *fh;
+
+  g_static_mutex_lock (&global_mutex);
+
+  fh = GET_FILE_HANDLE (fi);
+
+  /* If the file handle is still valid, its value won't change. If
+   * invalid, it's set to NULL. */
+  fh = g_hash_table_lookup (global_active_fh_map, fh);
+
+  if (fh)
+    file_handle_ref (fh);
+
+  g_static_mutex_unlock (&global_mutex);
   return fh;
 }
 
@@ -306,17 +357,17 @@ reindex_file_handle_for_path (const gchar *old_path, const gchar *new_path)
 
   g_static_mutex_lock (&global_mutex);
 
-  if (!g_hash_table_lookup_extended (global_fh_table, old_path,
+  if (!g_hash_table_lookup_extended (global_path_to_fh_map, old_path,
                                      (gpointer *) &old_path_internal,
                                      (gpointer *) &fh))
       goto out;
 
-  g_hash_table_steal (global_fh_table, old_path);
+  g_hash_table_steal (global_path_to_fh_map, old_path);
 
   g_free (fh->path);
   fh->path = g_strdup (new_path);
 
-  g_hash_table_insert (global_fh_table, fh->path, fh);
+  g_hash_table_insert (global_path_to_fh_map, fh->path, fh);
 
  out:
   g_static_mutex_unlock (&global_mutex);
@@ -638,7 +689,7 @@ vfs_statfs (const gchar *path, struct statvfs *stbuf)
       g_object_unref (file);
     }
 
-  debug_print ("vfs_statfs: -> %s\n", strerror (-result));
+  debug_print ("vfs_statfs: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -666,10 +717,7 @@ vfs_statfs (const gchar *path, struct statvfs *stbuf)
   stbuf->f_flag = 0;  /* Ignored by FUSE */
   stbuf->f_namemax = 1024;
 
-  if (statvfs ("/", stbuf) != 0)
-    result = -errno;
-
-  debug_print ("vfs_statfs: -> %s\n", strerror (-result));
+  debug_print ("vfs_statfs: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -726,7 +774,19 @@ getattr_for_file (GFile *file, struct stat *sbuf)
   GError    *error  = NULL;
   gint       result = 0;
 
-  file_info = g_file_query_info (file, "*", 0, NULL, &error);
+  file_info = g_file_query_info (file, 
+                                 G_FILE_ATTRIBUTE_STANDARD_TYPE ","
+                                 G_FILE_ATTRIBUTE_STANDARD_NAME ","
+                                 G_FILE_ATTRIBUTE_STANDARD_IS_SYMLINK ","
+                                 G_FILE_ATTRIBUTE_STANDARD_SIZE ","
+                                 G_FILE_ATTRIBUTE_UNIX_MODE ","
+                                 G_FILE_ATTRIBUTE_TIME_CHANGED ","
+                                 G_FILE_ATTRIBUTE_TIME_MODIFIED ","
+                                 G_FILE_ATTRIBUTE_TIME_ACCESS ","
+                                 G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE ","
+                                 G_FILE_ATTRIBUTE_UNIX_BLOCKS ","
+				 "access::*",
+                                 0, NULL, &error);
 
   if (file_info)
     {
@@ -747,6 +807,13 @@ getattr_for_file (GFile *file, struct stat *sbuf)
       if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_TIME_ACCESS))
         sbuf->st_atime = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_TIME_ACCESS);
 
+      if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE))
+	sbuf->st_blksize = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCK_SIZE);
+      if (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCKS))
+	sbuf->st_blocks = file_info_get_attribute_as_uint (file_info, G_FILE_ATTRIBUTE_UNIX_BLOCKS);
+      else /* fake it to make 'du' work like 'du --apparent'. */
+	sbuf->st_blocks = (sbuf->st_size + 511) / 512;
+
       /* Setting st_nlink to 1 for directories makes 'find' work */
       sbuf->st_nlink = 1;
 
@@ -770,19 +837,26 @@ getattr_for_file (GFile *file, struct stat *sbuf)
   return result;
 }
 
+static void
+getattr_for_file_handle (FileHandle *fh, struct stat *sbuf)
+{
+  sbuf->st_mode = S_IFREG | S_IRUSR | S_IWUSR | S_IXUSR;
+  sbuf->st_uid = daemon_uid;
+  sbuf->st_gid = daemon_gid;
+  sbuf->st_nlink = 1;
+  sbuf->st_size = fh->pos;
+  sbuf->st_blksize = 512;
+  sbuf->st_blocks = (sbuf->st_size + 511) / 512;
+}
+
 static gint
 vfs_getattr (const gchar *path, struct stat *sbuf)
 {
   GFile      *file;
   gint        result = 0;
-  FileHandle *fh;
 
   debug_print ("vfs_getattr: %s\n", path);
 
-  fh = get_file_handle_for_path (path);
-  if (fh)
-    g_mutex_lock (fh->mutex);
-
   memset (sbuf, 0, sizeof (*sbuf));
 
   sbuf->st_dev = 0;                     /* dev_t     ID of device containing file */
@@ -814,6 +888,26 @@ vfs_getattr (const gchar *path, struct stat *sbuf)
       /* Submount */
 
       result = getattr_for_file (file, sbuf);
+
+      if (result != 0)
+        {
+          FileHandle *fh = get_file_handle_for_path (path);
+
+          /* Some backends don't create new files until their stream has
+           * been closed. So, if the path doesn't exist, but we have a stream
+           * associated with it, pretend it's there. */
+
+          if (fh != NULL)
+            {
+              g_mutex_lock (fh->mutex);
+              getattr_for_file_handle (fh, sbuf);
+              g_mutex_unlock (fh->mutex);
+
+              file_handle_unref (fh);
+              result = 0;
+            }
+        }
+
       g_object_unref (file);
     }
   else
@@ -821,13 +915,7 @@ vfs_getattr (const gchar *path, struct stat *sbuf)
       result = -ENOENT;
     }
 
-  if (fh)
-    {
-      g_mutex_unlock (fh->mutex);
-      file_handle_unref (fh);
-    }
-
-  debug_print ("vfs_getattr: -> %s\n", strerror (-result));
+  debug_print ("vfs_getattr: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -837,7 +925,12 @@ vfs_readlink (const gchar *path, gchar *target, size_t size)
 {
   debug_print ("vfs_readlink: %s\n", path);
 
-  return 0;
+  /* This is not implemented because it would allow remote servers to launch
+   * symlink attacks on the local machine.  There's not much of a use for
+   * "readlink" anyway since we don't pass G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+   * so non-broken symlinks will be followed transparently. */
+
+  return -ENOSYS;
 }
 
 static gint
@@ -887,7 +980,7 @@ setup_input_stream (GFile *file, FileHandle *fh)
 }
 
 static gint
-setup_output_stream (GFile *file, FileHandle *fh)
+setup_output_stream (GFile *file, FileHandle *fh, int flags)
 {
   GError *error  = NULL;
   gint    result = 0;
@@ -907,7 +1000,10 @@ setup_output_stream (GFile *file, FileHandle *fh)
 
   if (!fh->stream)
     {
-      fh->stream = g_file_append_to (file, 0, NULL, &error);
+      if (flags & O_TRUNC)
+        fh->stream = g_file_replace (file, NULL, FALSE, 0, NULL, &error);
+      else
+        fh->stream = g_file_append_to (file, 0, NULL, &error);
       if (fh->stream)
         fh->pos = g_seekable_tell (G_SEEKABLE (fh->stream));
     }
@@ -926,6 +1022,32 @@ setup_output_stream (GFile *file, FileHandle *fh)
   return result;
 }
 
+static gint
+open_common (const gchar *path, struct fuse_file_info *fi, GFile *file, int output_flags)
+{
+  gint        result;
+  FileHandle *fh = get_or_create_file_handle_for_path (path);
+
+  g_mutex_lock (fh->mutex);
+
+  SET_FILE_HANDLE (fi, fh);
+
+  debug_print ("open_common: flags=%o\n", fi->flags);
+
+  /* Set up a stream here, so we can check for errors */
+  set_pid_for_file (file);
+
+  if (fi->flags & O_WRONLY || fi->flags & O_RDWR)
+    result = setup_output_stream (file, fh, fi->flags | output_flags);
+  else
+    result = setup_input_stream (file, fh);
+
+  g_mutex_unlock (fh->mutex);
+
+  /* The added reference to the file handle is released in vfs_release() */
+  return result;
+}
+
 static gint
 vfs_open (const gchar *path, struct fuse_file_info *fi)
 {
@@ -941,7 +1063,7 @@ vfs_open (const gchar *path, struct fuse_file_info *fi)
       GFileInfo *file_info;
       GError    *error = NULL;
 
-      file_info = g_file_query_info (file, "*", 0, NULL, &error);
+      file_info = g_file_query_info (file, G_FILE_ATTRIBUTE_STANDARD_TYPE, 0, NULL, &error);
 
       if (file_info)
         {
@@ -949,26 +1071,7 @@ vfs_open (const gchar *path, struct fuse_file_info *fi)
 
           if (file_type == G_FILE_TYPE_REGULAR)
             {
-              FileHandle *fh = get_or_create_file_handle_for_path (path);
-              
-              /* File exists */
-
-              SET_FILE_HANDLE (fi, fh);
-
-              debug_print ("vfs_open: flags=%o\n", fi->flags);
-
-              /* Set up a stream here, so we can check for errors */
-
-              g_mutex_lock (fh->mutex);
-
-              if (fi->flags & O_WRONLY || fi->flags & O_RDWR)
-                result = setup_output_stream (file, fh);
-              else
-                result = setup_input_stream (file, fh);
-
-              g_mutex_unlock (fh->mutex);
-
-              /* The added reference to the file handle is released in vfs_release() */
+              result = open_common (path, fi, file, 0);
             }
           else if (file_type == G_FILE_TYPE_DIRECTORY)
             {
@@ -1005,7 +1108,7 @@ vfs_open (const gchar *path, struct fuse_file_info *fi)
       result = -ENOENT;
     }
 
-  debug_print ("vfs_open: -> %s\n", strerror (-result));
+  debug_print ("vfs_open: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -1025,7 +1128,9 @@ vfs_create (const gchar *path, mode_t mode, struct fuse_file_info *fi)
       GFileInfo *file_info;
       GError    *error = NULL;
 
-      file_info = g_file_query_info (file, "*", 0, NULL, &error);
+      set_pid_for_file (file);
+
+      file_info = g_file_query_info (file, G_FILE_ATTRIBUTE_STANDARD_TYPE, 0, NULL, &error);
 
       if (file_info)
         {
@@ -1049,17 +1154,17 @@ vfs_create (const gchar *path, mode_t mode, struct fuse_file_info *fi)
 
               /* Success */
 
-              SET_FILE_HANDLE (fi, fh);
-
               g_mutex_lock (fh->mutex);
 
+              SET_FILE_HANDLE (fi, fh);
+
               file_handle_close_stream (fh);
               fh->stream = file_output_stream;
               fh->op = FILE_OP_WRITE;
 
               g_mutex_unlock (fh->mutex);
 
-              /* The added reference to the file handle is released in vfs_release() */
+              /* The reference added to the file handle is released in vfs_release() */
             }
           else
             {
@@ -1075,7 +1180,7 @@ vfs_create (const gchar *path, mode_t mode, struct fuse_file_info *fi)
       result = -ENOENT;
     }
 
-  debug_print ("vfs_create: -> %s\n", strerror (-result));
+  debug_print ("vfs_create: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -1083,12 +1188,16 @@ vfs_create (const gchar *path, mode_t mode, struct fuse_file_info *fi)
 static gint
 vfs_release (const gchar *path, struct fuse_file_info *fi)
 {
-  FileHandle *fh = GET_FILE_HANDLE (fi);
+  FileHandle *fh = get_file_handle_from_info (fi);
 
   debug_print ("vfs_release: %s\n", path);
 
   if (fh)
-    file_handle_unref (fh);
+    {
+      /* get_file_handle_from_info () adds a "working ref", so unref twice. */
+      file_handle_unref (fh);
+      file_handle_unref (fh);
+    }
 
   return 0;
 }
@@ -1205,23 +1314,32 @@ vfs_read (const gchar *path, gchar *buf, size_t size,
 
   if ((file = file_from_full_path (path)))
     {
-      FileHandle *fh = GET_FILE_HANDLE (fi);
+      FileHandle *fh = get_file_handle_from_info (fi);
 
-      g_mutex_lock (fh->mutex);
-
-      result = setup_input_stream (file, fh);
-
-      if (result == 0)
+      if (fh)
         {
-          result = read_stream (fh, buf, size, offset);
+          g_mutex_lock (fh->mutex);
+
+          result = setup_input_stream (file, fh);
+
+          if (result == 0)
+            {
+              result = read_stream (fh, buf, size, offset);
+            }
+          else
+            {
+              debug_print ("vfs_read: failed to setup input_stream!\n");
+            }
+
+          g_mutex_unlock (fh->mutex);
+          file_handle_unref (fh);
         }
       else
         {
-          debug_print ("vfs_read: failed to setup input_stream!\n");
+          result = -EINVAL;
         }
 
       g_object_unref (file);
-      g_mutex_unlock (fh->mutex);
     }
   else
     {
@@ -1229,7 +1347,7 @@ vfs_read (const gchar *path, gchar *buf, size_t size,
     }
 
   if (result < 0)
-    debug_print ("vfs_read: -> %s\n", strerror (-result));
+    debug_print ("vfs_read: -> %s\n", g_strerror (-result));
   else
     debug_print ("vfs_read: -> %d bytes read.\n", result);
 
@@ -1324,18 +1442,27 @@ vfs_write (const gchar *path, const gchar *buf, size_t len, off_t offset,
 
   if ((file = file_from_full_path (path)))
     {
-      FileHandle *fh = GET_FILE_HANDLE (fi);
+      FileHandle *fh = get_file_handle_from_info (fi);
 
-      g_mutex_lock (fh->mutex);
-
-      result = setup_output_stream (file, fh);
-      if (result == 0)
+      if (fh)
         {
-          result = write_stream (fh, buf, len, offset);
+          g_mutex_lock (fh->mutex);
+
+          result = setup_output_stream (file, fh, 0);
+          if (result == 0)
+            {
+              result = write_stream (fh, buf, len, offset);
+            }
+
+          g_mutex_unlock (fh->mutex);
+          file_handle_unref (fh);
+        }
+      else
+        {
+          result = -EINVAL;
         }
 
       g_object_unref (file);
-      g_mutex_unlock (fh->mutex);
     }
   else
     {
@@ -1343,7 +1470,7 @@ vfs_write (const gchar *path, const gchar *buf, size_t len, off_t offset,
     }
 
   if (result < 0)
-    debug_print ("vfs_write: -> %s\n", strerror (-result));
+    debug_print ("vfs_write: -> %s\n", g_strerror (-result));
   else
     debug_print ("vfs_write: -> %d bytes written.\n", result);
 
@@ -1353,8 +1480,42 @@ vfs_write (const gchar *path, const gchar *buf, size_t len, off_t offset,
 static gint
 vfs_flush (const gchar *path, struct fuse_file_info *fi)
 {
+  FileHandle *fh = get_file_handle_from_info (fi);
+
   debug_print ("vfs_flush: %s\n", path);
 
+  if (fh)
+    {
+      g_mutex_lock (fh->mutex);
+      file_handle_close_stream (fh);
+      g_mutex_unlock (fh->mutex);
+
+      /* get_file_handle_from_info () adds a "working ref", so release that. */
+      file_handle_unref (fh);
+    }
+
+  /* TODO: Error handling. */
+  return 0;
+}
+
+static gint
+vfs_fsync (const gchar *path, gint sync_data_only, struct fuse_file_info *fi)
+{
+  FileHandle *fh = get_file_handle_from_info (fi);
+
+  debug_print ("vfs_flush: %s\n", path);
+
+  if (fh)
+    {
+      g_mutex_lock (fh->mutex);
+      file_handle_close_stream (fh);
+      g_mutex_unlock (fh->mutex);
+
+      /* get_file_handle_from_info () adds a "working ref", so release that. */
+      file_handle_unref (fh);
+    }
+
+  /* TODO: Error handling. */
   return 0;
 }
 
@@ -1493,9 +1654,7 @@ vfs_rename (const gchar *old_path, const gchar *new_path)
 
   if (old_file && new_file)
     {
-      FileHandle *fh;
-
-      fh = get_file_handle_for_path (old_path);
+      FileHandle *fh = get_file_handle_for_path (old_path);
 
       if (fh)
         {
@@ -1539,7 +1698,7 @@ vfs_rename (const gchar *old_path, const gchar *new_path)
   if (new_file)
     g_object_unref (new_file);
 
-  debug_print ("vfs_rename: -> %s\n", strerror (-result));
+  debug_print ("vfs_rename: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -1557,9 +1716,7 @@ vfs_unlink (const gchar *path)
 
   if (file)
     {
-      FileHandle *fh;
-
-      fh = get_file_handle_for_path (path);
+      FileHandle *fh = get_file_handle_for_path (path);
 
       if (fh)
         {
@@ -1590,7 +1747,7 @@ vfs_unlink (const gchar *path)
       result = -ENOENT;
     }
 
-  debug_print ("vfs_unlink: -> %s\n", strerror (-result));
+  debug_print ("vfs_unlink: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -1628,7 +1785,7 @@ vfs_mkdir (const gchar *path, mode_t mode)
       result = -ENOENT;
     }
 
-  debug_print ("vfs_mkdir: -> %s\n", strerror (-result));
+  debug_print ("vfs_mkdir: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -1648,7 +1805,7 @@ vfs_rmdir (const gchar *path)
     {
       GFileInfo *file_info;
 
-      file_info = g_file_query_info (file, "*", 0, NULL, &error);
+      file_info = g_file_query_info (file, G_FILE_ATTRIBUTE_STANDARD_TYPE, 0, NULL, &error);
       if (file_info)
         {
           if (g_file_info_get_file_type (file_info) == G_FILE_TYPE_DIRECTORY)
@@ -1684,17 +1841,51 @@ vfs_rmdir (const gchar *path)
       result = -ENOENT;
     }
 
-  debug_print ("vfs_rmdir: -> %s\n", strerror (-result));
+  debug_print ("vfs_rmdir: -> %s\n", g_strerror (-result));
 
   return result;
 }
 
+static gboolean
+file_handle_get_size (FileHandle *fh,
+		      goffset *size)
+{
+  GFileInfo *info;
+  gboolean res;
+
+  if (fh->stream == NULL)
+    return FALSE;
+  
+  info = NULL;
+  if (fh->op == FILE_OP_READ)
+    info = g_file_input_stream_query_info (fh->stream,
+					   G_FILE_ATTRIBUTE_STANDARD_SIZE,
+					   NULL, NULL);
+  else if (fh->op == FILE_OP_WRITE)
+    info = g_file_output_stream_query_info (fh->stream,
+					    G_FILE_ATTRIBUTE_STANDARD_SIZE,
+					    NULL, NULL);
+
+  res = FALSE;
+  if (info)
+    {
+      if (g_file_info_has_attribute (info, G_FILE_ATTRIBUTE_STANDARD_SIZE))
+	{
+	  *size = g_file_info_get_size (info);
+	  res = TRUE;
+	}
+      g_object_unref (info);
+    }
+  return res;
+}
+
 static gint
 vfs_ftruncate (const gchar *path, off_t size, struct fuse_file_info *fi)
 {
   GFile  *file;
   GError *error  = NULL;
   gint    result = 0;
+  goffset current_size;
 
   debug_print ("vfs_ftruncate: %s\n", path);
 
@@ -1702,47 +1893,71 @@ vfs_ftruncate (const gchar *path, off_t size, struct fuse_file_info *fi)
 
   if (file)
     {
-      FileHandle *fh = GET_FILE_HANDLE (fi);
+      FileHandle *fh = get_file_handle_from_info (fi);
 
-      g_mutex_lock (fh->mutex);
-
-      result = setup_output_stream (file, fh);
-
-      if (result == 0)
+      if (fh)
         {
-          if (g_seekable_can_truncate (G_SEEKABLE (fh->stream)))
-            {
-              g_seekable_truncate (fh->stream, size, NULL, &error);
-            }
-          else if (size == 0)
-            {
-              g_output_stream_close (fh->stream, NULL, NULL);
-              g_object_unref (fh->stream);
-              fh->stream = NULL;
+          g_mutex_lock (fh->mutex);
 
-              fh->stream = g_file_replace (file, 0, FALSE, 0, NULL, &error);
-            }
-          else
+          result = setup_output_stream (file, fh, 0);
+
+          if (result == 0)
             {
-              result = -ENOTSUP;
+              if (g_seekable_can_truncate (G_SEEKABLE (fh->stream)))
+                {
+                  g_seekable_truncate (fh->stream, size, NULL, &error);
+                }
+              else if (size == 0)
+                {
+                  g_output_stream_close (fh->stream, NULL, NULL);
+                  g_object_unref (fh->stream);
+                  fh->stream = NULL;
+
+                  fh->stream = g_file_replace (file, 0, FALSE, 0, NULL, &error);
+
+                  if (fh->stream != NULL)
+                    {
+                      /* The stream created by g_file_replace() won't always replace
+                       * the file until it's been closed. So close it now to make
+                       * future operations consistent. */
+                      g_output_stream_close (fh->stream, NULL, NULL);
+                      g_object_unref (fh->stream);
+                      fh->stream = NULL;
+                    }
+                }
+              else if (file_handle_get_size (fh, &current_size) &&
+		       current_size == size)
+		{
+		  /* Don't have to do anything to succeed */
+		}
+	      else
+		{
+		  result = -ENOTSUP;
+                }
+
+              if (error)
+                {
+                  result = -errno_from_error (error);
+                  g_error_free (error);
+                }
             }
 
-          if (error)
-            {
-              result = -errno_from_error (error);
-              g_error_free (error);
-            }
+          g_mutex_unlock (fh->mutex);
+          file_handle_unref (fh);
+        }
+      else
+        {
+          result = -EINVAL;
         }
 
       g_object_unref (file);
-      g_mutex_unlock (fh->mutex);
     }
   else
     {
       result = -ENOENT;
     }
 
-  debug_print ("vfs_ftruncate: -> %s\n", strerror (-result));
+  debug_print ("vfs_ftruncate: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -1763,6 +1978,7 @@ vfs_truncate (const gchar *path, off_t size)
       GFileOutputStream *file_output_stream = NULL;
       FileHandle        *fh;
 
+      /* Get a file handle just to lock the path while we're working */
       fh = get_file_handle_for_path (path);
       if (fh)
         g_mutex_lock (fh->mutex);
@@ -1803,7 +2019,7 @@ vfs_truncate (const gchar *path, off_t size)
       result = -ENOENT;
     }
 
-  debug_print ("vfs_truncate: -> %s\n", strerror (-result));
+  debug_print ("vfs_truncate: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -1834,7 +2050,7 @@ vfs_symlink (const gchar *path_old, const gchar *path_new)
       result = -ENOENT;
     }
 
-  debug_print ("vfs_symlink: -> %s\n", strerror (-result));
+  debug_print ("vfs_symlink: -> %s\n", g_strerror (-result));
 
   return result;
 }
@@ -1854,7 +2070,11 @@ vfs_access (const gchar *path, gint mode)
     {
       GFileInfo *file_info;
 
-      file_info = g_file_query_info (file, "*", 0, NULL, &error);
+      file_info = g_file_query_info (file,
+                                     G_FILE_ATTRIBUTE_ACCESS_CAN_READ ","
+                                     G_FILE_ATTRIBUTE_ACCESS_CAN_WRITE ","
+                                     G_FILE_ATTRIBUTE_ACCESS_CAN_EXECUTE ",",
+                                     0, NULL, &error);
       if (file_info)
         {
           if ((mode & R_OK && (g_file_info_has_attribute (file_info, G_FILE_ATTRIBUTE_ACCESS_CAN_READ) &&
@@ -1877,6 +2097,23 @@ vfs_access (const gchar *path, gint mode)
           result = -EIO;
         }
 
+      if (result != 0)
+        {
+          FileHandle *fh = get_file_handle_for_path (path);
+
+          /* Some backends don't create new files until their stream has
+           * been closed. So, if the path doesn't exist, but we have a stream
+           * associated with it, pretend it's there. */
+
+          if (fh != NULL)
+            {
+              file_handle_unref (fh);
+
+              /* The file was presumably created by us, so indicate full access. */
+              result = 0;
+            }
+        }
+
       g_object_unref (file);
     }
   else if (path_is_mount_list (path))
@@ -1888,7 +2125,7 @@ vfs_access (const gchar *path, gint mode)
       result = -ENOENT;
     }
 
-  debug_print ("vfs_access: -> %s\n", strerror (-result));
+  debug_print ("vfs_access: -> %s\n", g_strerror (-result));
   return result;
 }
 
@@ -1906,40 +2143,53 @@ vfs_utimens (const gchar *path, const struct timespec tv [2])
   if (file)
     {
       guint64 atime;
+      guint32 atime_usec;
       guint64 mtime;
+      guint32 mtime_usec;
+      GFileInfo *info;
 
       if (tv)
         {
-          atime = (guint64) tv [0].tv_sec * 1000000 + (guint64) tv [0].tv_nsec / (guint64) 1000;
-          mtime = (guint64) tv [1].tv_sec * 1000000 + (guint64) tv [1].tv_nsec / (guint64) 1000;
+          atime = (guint64) tv [0].tv_sec;
+          atime_usec = (guint32) tv [0].tv_nsec / (guint32) 1000;
+          mtime = (guint64) tv [1].tv_sec;
+          mtime_usec = (guint32) tv [1].tv_nsec / (guint32) 1000;
         }
       else
         {
           struct timeval tiv;
 
           gettimeofday (&tiv, NULL);
-          atime = (guint64) tiv.tv_sec * (guint64) 1000000 + (guint64) tiv.tv_usec;
+          atime = (guint64) tiv.tv_sec;
+          atime_usec = (guint32) tiv.tv_usec;
           mtime = atime;
+          mtime_usec = atime_usec;
         }
 
-      g_file_set_attribute (file, G_FILE_ATTRIBUTE_TIME_ACCESS_USEC,
-                            G_FILE_ATTRIBUTE_TYPE_UINT64, &atime,
-                            0, NULL, &error);
+      info = g_file_info_new ();
+      g_file_info_set_attribute_uint64 (info, G_FILE_ATTRIBUTE_TIME_MODIFIED, mtime);
+      g_file_info_set_attribute_uint32 (info, G_FILE_ATTRIBUTE_TIME_MODIFIED_USEC, mtime_usec);
+      g_file_info_set_attribute_uint64 (info, G_FILE_ATTRIBUTE_TIME_ACCESS, atime);
+      g_file_info_set_attribute_uint32 (info, G_FILE_ATTRIBUTE_TIME_ACCESS_USEC, atime_usec);
 
-      if (!error)
-        {
-          g_file_set_attribute (file, G_FILE_ATTRIBUTE_TIME_MODIFIED_USEC,
-                                G_FILE_ATTRIBUTE_TYPE_UINT64, &mtime,
-                                0, NULL, &error);
-        }
+      g_file_set_attributes_from_info (file, info, 0, NULL, &error);
 
       if (error)
         {
-          result = -errno_from_error (error);
+          /* As long as not all backends support all attributes we set,
+             report failure only if neither mtime and atime have been set. */
+          if (g_file_info_get_attribute_status (info, G_FILE_ATTRIBUTE_TIME_ACCESS) == G_FILE_ATTRIBUTE_STATUS_ERROR_SETTING &&
+              g_file_info_get_attribute_status (info, G_FILE_ATTRIBUTE_TIME_MODIFIED) == G_FILE_ATTRIBUTE_STATUS_ERROR_SETTING)
+            {
+              /* Note: we only get first error from the attributes we try to set,  might not be accurate
+                       (a limitation of g_file_set_attributes_from_info()). */
+              result = -errno_from_error (error);
+            }
           g_error_free (error);
         }
 
       g_object_unref (file);
+      g_object_unref (info);
     }
   else if (path_is_mount_list (path))
     {
@@ -1950,7 +2200,7 @@ vfs_utimens (const gchar *path, const struct timespec tv [2])
       result = -ENOENT;
     }
 
-  debug_print ("vfs_utimens: -> %s\n", strerror (-result));
+  debug_print ("vfs_utimens: -> %s\n", g_strerror (-result));
   return result;
 }
 
@@ -2003,15 +2253,13 @@ mount_tracker_unmounted_cb (GVolumeMonitor *volume_monitor,
   GList *l;
 
   root = g_mount_get_root (mount);
-  
+
   mount_list_lock ();
-  
-  root = g_mount_get_root (mount);
 
   for (l = mount_list; l != NULL; l = l->next)
     {
       MountRecord *mount_record = l->data;
-      
+
       if (g_file_equal (root, mount_record->root))
         {
           mount_list = g_list_delete_link (mount_list, l);
@@ -2019,7 +2267,7 @@ mount_tracker_unmounted_cb (GVolumeMonitor *volume_monitor,
           break;
         }
     }
-      
+
   mount_list_unlock ();
 
   g_object_unref (root);
@@ -2040,6 +2288,7 @@ subthread_main (gpointer data)
   g_signal_handlers_disconnect_by_func (volume_monitor, mount_tracker_unmounted_cb, NULL);
 
   g_main_loop_unref (subthread_main_loop);
+  subthread_main_loop = NULL;
 
   g_object_unref (volume_monitor);
   volume_monitor = NULL;
@@ -2055,7 +2304,7 @@ dbus_filter_func (DBusConnection *connection,
                   DBusMessage    *message,
                   void           *data)
 {
-	if (dbus_message_is_signal (message,
+  if (dbus_message_is_signal (message,
                               DBUS_INTERFACE_DBUS,
                               "NameOwnerChanged"))
     {
@@ -2084,8 +2333,8 @@ dbus_filter_func (DBusConnection *connection,
       /* Session bus died, unmount */
       g_main_loop_quit (subthread_main_loop);
     }
-	
-	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
+  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
 static gpointer
@@ -2100,12 +2349,14 @@ vfs_init (struct fuse_conn_info *conn)
   daemon_gid = getgid ();
 
   mount_list_mutex = g_mutex_new ();
-  global_fh_table = g_hash_table_new_full (g_str_hash, g_str_equal,
-                                           NULL, (GDestroyNotify) file_handle_free);
+  global_path_to_fh_map = g_hash_table_new_full (g_str_hash, g_str_equal,
+                                                 NULL, (GDestroyNotify) file_handle_free);
+  global_active_fh_map = g_hash_table_new_full (g_direct_hash, g_direct_equal,
+                                                NULL, NULL);
 
-	dbus_error_init (&error);
+  dbus_error_init (&error);
 
-	dbus_conn = dbus_bus_get (DBUS_BUS_SESSION, &error);
+  dbus_conn = dbus_bus_get (DBUS_BUS_SESSION, &error);
   if (dbus_conn == NULL)
     {
       g_printerr ("Failed to connect to the D-BUS daemon: %s\n",
@@ -2118,13 +2369,13 @@ vfs_init (struct fuse_conn_info *conn)
 
   _g_dbus_connection_integrate_with_main (dbus_conn);
 
-	dbus_bus_add_match (dbus_conn,
+  dbus_bus_add_match (dbus_conn,
                       "type='signal',sender='" DBUS_SERVICE_DBUS "',"
                       "interface='" DBUS_INTERFACE_DBUS "',"
                       "member='NameOwnerChanged',"
                       "arg0='"G_VFS_DBUS_DAEMON_NAME"'",
                       NULL);
-	dbus_connection_add_filter (dbus_conn,
+  dbus_connection_add_filter (dbus_conn,
                               dbus_filter_func,
                               NULL,
                               NULL);
@@ -2144,6 +2395,9 @@ vfs_init (struct fuse_conn_info *conn)
   subthread_main_loop = g_main_loop_new (NULL, FALSE);
   subthread = g_thread_create ((GThreadFunc) subthread_main, NULL, FALSE, NULL);
 
+  /* Indicate O_TRUNC support for open() */
+  conn->want |= FUSE_CAP_ATOMIC_O_TRUNC;
+
   return NULL;
 }
 
@@ -2151,7 +2405,8 @@ static void
 vfs_destroy (gpointer param)
 {
   mount_list_free ();
-  g_main_loop_quit (subthread_main_loop);
+  if (subthread_main_loop != NULL) 
+    g_main_loop_quit (subthread_main_loop);
   g_mutex_free (mount_list_mutex);
   g_object_unref (gvfs);
 }
@@ -2174,6 +2429,7 @@ static struct fuse_operations vfs_oper =
   .create      = vfs_create,
   .release     = vfs_release,
   .flush       = vfs_flush,
+  .fsync       = vfs_fsync,
 
   .read        = vfs_read,
   .write       = vfs_write,
