@@ -58,7 +58,7 @@
 #include <gvfsjobpull.h>
 #include <gvfsjobsetattribute.h>
 #include <gvfsjobqueryattributes.h>
-#include <gdbusutils.h>
+#include <gvfsdbusutils.h>
 
 enum {
   PROP_0,
@@ -80,6 +80,7 @@ struct _GVfsBackendPrivate
   gboolean user_visible;
   char *default_location;
   GMountSpec *mount_spec;
+  gboolean block_requests;
 };
 
 
@@ -509,6 +510,19 @@ g_vfs_backend_add_auto_info (GVfsBackend *backend,
   
 }
 
+void
+g_vfs_backend_set_block_requests (GVfsBackend *backend)
+{
+  backend->priv->block_requests = TRUE;
+}
+
+gboolean
+g_vfs_backend_get_block_requests (GVfsBackend *backend)
+{
+  return backend->priv->block_requests;
+}
+
+
 static DBusHandlerResult
 backend_dbus_handler (DBusConnection  *connection,
 		      DBusMessage     *message,
@@ -522,6 +536,21 @@ backend_dbus_handler (DBusConnection  *connection,
   g_debug ("backend_dbus_handler %s:%s\n",
 	   dbus_message_get_interface (message),
 	   dbus_message_get_member (message));
+
+  if (backend->priv->block_requests)
+    {
+      DBusMessage *reply;
+
+      reply = _dbus_message_new_gerror (message,
+					G_IO_ERROR,
+					G_IO_ERROR_NOT_MOUNTED,
+					"%s", "Backend currently unmounting");
+
+      g_assert (reply != NULL);
+
+      dbus_connection_send (connection, reply, NULL);
+      return DBUS_HANDLER_RESULT_HANDLED;
+    }
   
   if (dbus_message_is_method_call (message,
 				   G_VFS_DBUS_MOUNT_INTERFACE,
