@@ -1113,3 +1113,19 @@ g_vfs_daemon_get_blocking_processes (GVfsDaemon *daemon)
   return processes;
 }
 
+void
+g_vfs_daemon_run_job_in_thread (GVfsDaemon *daemon,
+				GVfsJob    *job)
+{
+  g_thread_pool_push (daemon->thread_pool, job, NULL); /* TODO: Check error */
+}
+
+void
+g_vfs_daemon_close_active_channels (GVfsDaemon *daemon)
+{
+  GList *l;
+
+   for (l = daemon->job_sources; l != NULL; l = l->next)
+      if (G_VFS_IS_CHANNEL (l->data))
+        g_vfs_channel_force_close (G_VFS_CHANNEL (l->data));
+}
